/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml;

import com.ssg.tools.jsonxml.utils.Path;
import com.ssg.tools.jsonxml.utils.StructureProcessingContext;
import com.ssg.tools.jsonxml.utils.Utilities;
import com.ssg.tools.jsonxml.utils.Utilities.Pair;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Compares 2 structures providing delta(s).
 * 
 * @author ssg
 */
public class Comparer {

    public enum COMPARE_STATUS {

        unknown, match, mismatch
    }

    public enum MATCH_QUALITY {

        no, full, greater, less, value_match, value_mismatch, type_mismatch, missing_a, missing_b, nested_mismatch
    }

    /**
     * Encapsulates comparison results for pair of objects.
     * Comparison results are from the point of view of object A.
     * 
     * For extended comparison strategies object B may have alternative path.
     */
    public static class ComparatorPair extends Pair {

        private COMPARE_STATUS _status = COMPARE_STATUS.unknown;
        private MATCH_QUALITY _quality = MATCH_QUALITY.no;
        private List<ComparatorPair> _nested = new ArrayList<ComparatorPair>();
        private String _path;
        private String _pathB;
        private boolean _multiMatch = false;

        public ComparatorPair(Object a, Object b) {
            super(a, b);
        }

        public ComparatorPair(String path, Object a, Object b) {
            super(a, b);
            _path = path;
        }

        /**
         * @return the _status
         */
        public COMPARE_STATUS getStatus() {
            return _status;
        }

        /**
         * @param status the _status to set
         */
        public void setStatus(COMPARE_STATUS status) {
            this._status = status;
        }

        /**
         * @return the _nested
         */
        public List<ComparatorPair> getNested() {
            return _nested;
        }

        public void addNested(ComparatorPair cp) {
            if (cp != null && !_nested.contains(cp)) {
                _nested.add(cp);
            }
        }

        /**
         * @return the _path
         */
        public String getPath() {
            return _path;
        }

        /**
         * @param path the _path to set
         */
        public void setPath(String path) {
            this._path = path;
        }

        /**
         * @return the _quality
         */
        public MATCH_QUALITY getQuality() {
            return _quality;
        }

        /**
         * @param quality the _quality to set
         */
        public void setQuality(MATCH_QUALITY quality) {
            this._quality = quality;
        }

        /**
         * @return the _multiMatch
         */
        public boolean isMultiMatch() {
            return _multiMatch;
        }

        /**
         * @param multiMatch the _multiMatch to set
         */
        public void setMultiMatch(boolean multiMatch) {
            this._multiMatch = multiMatch;
        }

        /**
         * @return the _pathB
         */
        public String getPathB() {
            return (_pathB != null) ? _pathB : _path;
        }

        /**
         * @param pathB the _pathB to set
         */
        public void setPathB(String pathB) {
            this._pathB = pathB;
        }

        /**
         * Recursive search for matching pair.
         * Parameter pathB is optional.
         * 
         * @param path
         * @param pathB
         * @return 
         */
        public ComparatorPair find(String path, String pathB) {
            if (getPath().equals(path) && (pathB == null || getPathB().equals(pathB))) {
                return this;
            }
            if (path.startsWith(getPath()) && getNested() != null) {
                for (ComparatorPair scp : getNested()) {
                    scp = scp.find(path, pathB);
                    if (scp != null) {
                        return scp;
                    }
                }
            }
            return null;
        }

        /**
         * Recursive search for matching pair.
         * Parameter pathB is optional.
         * 
         * @param path
         * @param pathB
         * @return 
         */
        public ComparatorPair find(String path, Object a, Object b) {
            if (getPath().equals(path)) {
                boolean matchA = (getA() == null && a == null) || (getA() != null && getA().equals(a));
                boolean matchB = (getB() == null && b == null) || (getB() != null && getB().equals(b));
                if (matchA && matchB) {
                    return this;
                }
            }
            if (path.startsWith(getPath()) && getNested() != null) {
                for (ComparatorPair scp : getNested()) {
                    scp = scp.find(path, a, b);
                    if (scp != null) {
                        return scp;
                    }
                }
            }
            return null;
        }

        /**
         * Check if object A is same as provided. Use identity (System.identityHash()) if identityMath=true;
         * 
         * @param obj
         * @param identityMatch
         * @return 
         */
        public boolean matchA(Object obj, boolean identityMatch) {
            if (getA() == null && obj == null) {
                return true;
            }
            if (getA() != null && obj == null || getA() == null && obj != null) {
                return false;
            }
            if (identityMatch) {
                return System.identityHashCode(getA()) == System.identityHashCode(obj);
            } else {
                return getA().equals(obj);
            }
        }

        /**
         * Check if object B is same as provided. Use identity (System.identityHash()) if identityMath=true;
         * 
         * @param obj
         * @param identityMatch
         * @return 
         */
        public boolean matchB(Object obj, boolean identityMatch) {
            if (getB() == null && obj == null) {
                return true;
            }
            if (getB() != null && obj == null || getB() == null && obj != null) {
                return false;
            }
            if (identityMatch) {
                return System.identityHashCode(getB()) == System.identityHashCode(obj);
            } else {
                return getB().equals(obj);
            }
        }

        public static List<ComparatorPair> toList(ComparatorPair root, List<ComparatorPair> list) {
            if (list == null) {
                list = new ArrayList<ComparatorPair>();
            }
            list.add(root);
            for (ComparatorPair cp : root.getNested()) {
                toList(cp, list);
            }
            return list;
        }
    }

    /**
     * Provides generic context for processing and process control switches.
     * 
     *  skipObjects - allows to include or exclude references to compared objects from result. Default - false (objects included).
     *  unorderedLists - treat lists as unordered collections, i.e. search for matching items thru collection instead of using its index. Default - false (i.e. use index-based comparison).
     *  checkMultiMatch - in unordered lists enable search for next match after 1st found. Default - false (stop searching after 1st matching item).
     */
    public static class ComparatorContext extends StructureProcessingContext {

        private boolean _skipObjects = false;
        private boolean _unorderedLists = false;
        private boolean _checkMultiMatch = false;
        private boolean _reflectionAllowed = true;

        /**
         * @return the _skipObjects
         */
        public boolean isSkipObjects() {
            return _skipObjects;
        }

        /**
         * @param skipObjects the _skipObjects to set
         */
        public void setSkipObjects(boolean skipObjects) {
            this._skipObjects = skipObjects;
        }

        /**
         * @return the _unorderedLists
         */
        public boolean isUnorderedLists() {
            return _unorderedLists;
        }

        /**
         * @param unorderedLists the _unorderedLists to set
         */
        public void setUnorderedLists(boolean unorderedLists) {
            this._unorderedLists = unorderedLists;
        }

        /**
         * @return the _checkMultiMatch
         */
        public boolean isCheckMultiMatch() {
            return _checkMultiMatch;
        }

        /**
         * @param checkMultiMatch the _checkMultiMatch to set
         */
        public void setCheckMultiMatch(boolean checkMultiMatch) {
            this._checkMultiMatch = checkMultiMatch;
        }

        /**
         * @return the _reflectionAllowed
         */
        public boolean isReflectionAllowed() {
            return _reflectionAllowed;
        }

        /**
         * @param reflectionAllowed the _reflectionAllowed to set
         */
        public void setReflectionAllowed(boolean reflectionAllowed) {
            this._reflectionAllowed = reflectionAllowed;
        }
    }

    public ComparatorPair compare(ComparatorContext context, Object a, Object b, int depth) {
        return compare(context, a, b, depth, false);
    }

    public ComparatorPair compare(ComparatorContext context, Object a, Object b, int depth, boolean fakeObject) {
        //System.out.println("compare: START: " + context.getPath() + ", " + depth + ", " + fakeObject + ": " + a + "; " + b);
        if (context == null || depth < 0) {
            return null;
        }
        //
        if (context.getStack().contains(a)) {
            return null;
        }
        // add to stack and register in processed
        if (!fakeObject) {
            context.getStack().push(a);
            context.getObjectsRegistry().registerProcessedObject(context.getPath().toString(), a);
        }

        ComparatorPair cp = new ComparatorPair(context.getPath().toString(), a, b);
        if (context.isSkipObjects()) {
            cp.setA(null);
            cp.setB(null);
        }
        cp.setStatus(COMPARE_STATUS.match);
        cp.setQuality(MATCH_QUALITY.full);
        if (a == null && b == null) {
            // use default: all matches...
            //System.out.println("compare:      a=b=null");
        } else if (a != null && b == null) {
            cp.setStatus(COMPARE_STATUS.mismatch);
            cp.setQuality(MATCH_QUALITY.missing_b);
            //System.out.println("compare:      b=null");
        } else if (a == null) {
            cp.setStatus(COMPARE_STATUS.mismatch);
            cp.setQuality(MATCH_QUALITY.missing_a);
            //System.out.println("compare:      a=null");
        } else {
            if (a instanceof Map) {
                if (!(b instanceof Map) && !context.isReflectionAllowed()) {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.type_mismatch);
                    //System.out.println("compare:      a=map, b!=map");
                } else {
                    //System.out.println("compare:      a=map,b=map");
                    Map<String, Object> am = context.getReflectiveBuilder().wrapMap((Map) a);
                    Map<String, Object> bm = (b instanceof Map) ? context.getReflectiveBuilder().wrapMap((Map) b) : context.getReflectiveBuilder().getReflectable(b);
                    compareMaps(cp, context, am, bm, depth);
                }
            } else if (a instanceof List || a.getClass().isArray()) {
                if (b instanceof List || b.getClass().isArray()) {
                    List al = (a instanceof List) ? (List) a : Arrays.asList((Object[]) a);
                    List bl = (b instanceof List) ? (List) b : Arrays.asList((Object[]) b);
                    //System.out.println("compare:      a=list,b=list");
                    compareLists(cp, context, al, bl, depth);
                } else {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.type_mismatch);
                    //System.out.println("compare:      a=list,b!=list");
                }
            } else if (a.equals(b)) {
                cp.setStatus(COMPARE_STATUS.match);
                //System.out.println("compare:      a===b");
            } else if ((a.getClass().isAssignableFrom(b.getClass()) || b.getClass().isAssignableFrom(a.getClass()))
                    && a instanceof Comparable
                    && b instanceof Comparable) {
                int test = ((Comparable) a).compareTo(b);
                //System.out.println("compare:      a and b - comparable/assignable");
                //System.out.println("Compare comparables = "+test+" for: "+a+"; "+b);
                if (test < 0) {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.less);
                } else if (test > 0) {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.greater);
                }
            } else if (Utilities.isNumericClass(a.getClass()) && Utilities.isNumericClass(b.getClass())) {
                //System.out.println("compare:      a and b - numeric");
                Number an = (Number) a;
                Number bn = (Number) b;
                int test = ((Comparable) an.doubleValue()).compareTo(bn.doubleValue());
                //System.out.println("Compare numerics = "+test+" for: "+a+"; "+b);
                if (test < 0) {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.less);
                } else if (test > 0) {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.greater);
                }
            } else {
                if (!Utilities.isScalarOrString(a)) {
                    // reflective: convert object to map
                    //System.out.println("compare:      a=reflective Map");
                    Map map = context.getReflectiveBuilder().getReflectable(a);
                    cp = compare(context, map, b, depth, true);
                } else if (a.getClass().isAssignableFrom(b.getClass())) {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.value_match);
                } else {
                    cp.setStatus(COMPARE_STATUS.mismatch);
                    cp.setQuality(MATCH_QUALITY.type_mismatch);
                }
            }
        }
        if (!fakeObject) {
            context.getStack().pop();
        }
        //System.out.println("compare: END:   " + context.getPath() + ", " + depth + ": " + cp.getStatus());
        return cp;
    }

    void compareMaps(ComparatorPair cp, ComparatorContext context, Map<String, Object> a, Map<String, Object> b, int depth) {
        Map<String, Object> processedKeys = new HashMap<String, Object>();

        boolean matches = true;
        // process for a
        for (String key : a.keySet()) {
            processedKeys.put(key, key);
            context.getPath().add(key);
            ComparatorPair scp = compare(context, a.get(key), b.get(key), depth - 1, false);
            cp.addNested(scp);
            if (scp.getStatus() != COMPARE_STATUS.match) {
                matches = false;
            }
            context.getPath().remove();
        }
        // process missing for b
        for (String key : b.keySet()) {
            if (!processedKeys.containsKey(key)) {
                if (matches) {
                    matches = false;
                }
                processedKeys.put(key, key);
                context.getPath().add(key);
                ComparatorPair scp = compare(context, null, b.get(key), depth - 1, false);
                cp.addNested(scp);
                context.getPath().remove();
            }
        }
        if (!matches) {
            cp.setStatus(COMPARE_STATUS.mismatch);
            cp.setQuality(MATCH_QUALITY.nested_mismatch);
        }
    }

    void compareLists(ComparatorPair cp, ComparatorContext context, List a, List b, int depth) {
        boolean matches = true;
        if (context.isUnorderedLists()) {
            List matchingBLs = new ArrayList();
            for (int i = 0; i < a.size(); i++) {
                Object oa = a.get(i);
                // TODO: iterate thru all b items until 1st match...
                boolean matchesFound = false;
                for (int j = 0; j < b.size(); j++) {
                    Object ob = b.get(j);
                    int obID = System.identityHashCode(ob);
                    context.getPath().setIndex(i, true);
                    ComparatorPair scp = compare(context, oa, ob, depth - 1, false);
                    if (i != j) {
                        Path p = new Path(null);
                        p.copyPath(context.getPath());
                        p.setIndex(j, true);
                        scp.setPathB(p.toString());
                    }
                    if (scp.getStatus() == COMPARE_STATUS.match) {
                        cp.addNested(scp);
                        matchingBLs.add(obID);
                        if (!context.isCheckMultiMatch()) {
                            break;
                        }
                        matchesFound = true;
                    }
                }
                if (!matchesFound) {
                    matches = false;
                }
            }
            // insert extra items from list B
            for (int i = 0; i < b.size(); i++) {
                Object ob = b.get(i);
                // check if object was already matched - skip those.
                int obID = System.identityHashCode(ob);
                if (matchingBLs.contains(obID)) {
                    continue;
                }

                if (matches) {
                    matches = false;
                }
                context.getPath().setIndex(i, true);
                ComparatorPair scp = compare(context, null, ob, depth - 1);
                cp.addNested(scp);
            }
        } else {
            for (int i = 0; i < a.size(); i++) {
                Object oa = a.get(i);
                Object ob = (i < b.size()) ? b.get(i) : null;
                context.getPath().setIndex(i, true);
                ComparatorPair scp = compare(context, oa, ob, depth - 1, false);
                if (scp.getStatus() != COMPARE_STATUS.match) {
                    matches = false;
                }
                cp.addNested(scp);
            }
            // insert comparison for extra B items in list
            for (int i = a.size(); i < b.size(); i++) {
                if (matches) {
                    matches = false;
                }
                Object ob = (i < b.size()) ? b.get(i) : null;
                context.getPath().setIndex(i, true);
                ComparatorPair scp = compare(context, null, ob, depth - 1, false);
                cp.addNested(scp);
            }
        }
        context.getPath().setIndex(null, true);
        if (!matches) {
            cp.setStatus(COMPARE_STATUS.mismatch);
            cp.setQuality(MATCH_QUALITY.nested_mismatch);
        }

    }

    public String dumpComparatorPair(ComparatorPair cp, String prefix) {
        if (prefix == null) {
            prefix = "";
        }
        String delimiter = " ";
        StringBuilder sb = new StringBuilder(prefix);
        sb.append(cp.getPath());
        if (!cp.getPath().equals(cp.getPathB())) {
            sb.append(" -> ");
            sb.append(cp.getPathB());
        }
        sb.append(delimiter);
        sb.append(cp.getStatus());
        sb.append(':');
        sb.append(cp.getQuality());
        sb.append(delimiter);
        if (!cp.getNested().isEmpty()) {
            sb.append('\n');
            for (ComparatorPair scp : cp.getNested()) {
                if (scp != null) {
                    sb.append(dumpComparatorPair(scp, prefix + "    "));
                }
            }
        } else {
            sb.append('\n');
        }
        return sb.toString();
    }
}
