/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml;

import com.ssg.tools.jsonxml.impl.XMLFormat;
import com.ssg.tools.jsonxml.impl.StructureFormat;
import com.ssg.tools.jsonxml.impl.JSONFormat;
import com.ssg.tools.jsonxml.impl.StructureWriter;
import com.ssg.tools.jsonxml.utils.STRUCTURE_FORMAT_ELEMENT;
import com.ssg.tools.jsonxml.utils.PRIMITIVE_TYPE;
import com.ssg.tools.jsonxml.utils.ReflectiveBuilder;
import com.ssg.tools.jsonxml.utils.Utilities.Pair;
import java.io.Writer;
import java.io.IOException;
import java.io.StringWriter;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Alternative JSON/XML/etc formatter.
 * 
 * Based on Map/List/scalar data typing.
 * Unknown objects (not Map or List or known scalars) are converted into Map<String,Object> using getters.
 * Arrays are converted into Lists.
 * 
 * Uses StructureFormat to keep track of process and Formats instance in it to write locale-specific formatted data.
 * Structure format is used to construct in output lists, objects, their names, and values.
 * 
 * Default implementations of formatters are:
 * - StructureFormat: base implementation. Provides defaults and simple textual presentation of objects.
 * - DumpFormat: used to see object structure with data types.
 * - JSONFormat: simple JSON formatter.
 * - XMLFormat: simple XML formatter.
 * - XMLFormatForJSON1: modified XML formatter (adds attribute @json='list' for Lists/Arrays in original objects. (See XMLParser for reasoning)
 * 
 * JSON and XML formatting are explicitly exposed via static methods for usability and as example!
 * 
 * @author ssg
 *
 */
public abstract class Formatter {

    static Logger log = Logger.getLogger(Formatter.class.getName());

    public static String toJSONString(Object jo, boolean indented) {
        return toJSONString(null, jo, indented);
    }

    /**
     * Simple formatting call.
     * 
     * @param name
     * @param jo
     * @param indented
     * @return
     */
    public static String toJSONString(String name, Object jo, boolean indented) {
        return toJSONString(name, jo, new JSONFormat(indented));
    }

    public static String toJSONString(Object jo, JSONFormat format) {
        return toJSONString(null, jo, format);
    }

    /**
     * Simple formatting call.
     * 
     * @param name
     * @param jo
     * @param format
     * @return
     */
    public static String toJSONString(String name, Object jo, JSONFormat format) {
        StringWriter sw = new StringWriter();
        if (format == null) {
            format = new JSONFormat(false);
        }
        try {
            toFormattedString(sw, ensureObjectName(jo, name), format.getPrefix(), format, false);
        } catch (IOException ioex) {
            // TODO: nothing
        }
        return sw.toString();
    }

    public static String toXMLString(Object jo, boolean indented) {
        return toXMLString(null, jo, indented);
    }

    /**
     * Simple formatting call.
     * 
     * @param name
     * @param jo
     * @param indented
     * @return
     */
    public static String toXMLString(String name, Object jo, boolean indented) {
        return toXMLString(name, jo, new XMLFormat(indented));
    }

    public static String toXMLString(Object jo, XMLFormat format) {
        return toXMLString(null, jo, format);
    }

    /**
     * Simple formatting call.
     *
     * @param name
     * @param jo
     * @param format
     * @return
     */
    public static String toXMLString(String name, Object jo, XMLFormat format) {
        StringWriter sw = new StringWriter();
        if (format == null) {
            format = new XMLFormat(false);
        }
        try {
            toFormattedString(sw, ensureObjectName(jo, name), format.getPrefix(), format, false);
        } catch (IOException ioex) {
            // TODO: nothing
        }
        return sw.toString();
    }

    public static String toFormattedString(Object jo, StructureFormat format) {
        return toFormattedString(null, jo, format);
    }

    /**
     * Simple formatting call.
     * Use specified name to wrap result.
     *
     * @param name
     * @param jo
     * @param format
     * @return
     */
    public static String toFormattedString(String name, Object jo, StructureFormat format) {
        StringWriter sw = new StringWriter();
        if (format == null) {
            format = new StructureFormat(false);
        }
        try {
            toFormattedString(sw, ensureObjectName(jo, name), format.getPrefix(), format, false);
        } catch (IOException ioex) {
            // TODO: nothing
        }
        return sw.toString();
    }

    public static void toFormattedString(Writer sw, Object jo, String prefix, StructureFormat format, boolean fakeObject) throws IOException {
        if (!(sw instanceof StructureWriter)) {
            sw = new StructureWriter(sw, format);
        }
        toFormattedString((StructureWriter) sw, jo, prefix, format, fakeObject);
    }

    /**
     * Main recursive formatter method.
     * Handled specially numeric and boolean fields.
     *
     * NOTE: parameter fakeObject is used to indicate situations when real object is replaced with "convinient".
     * This is used when processing arrays as lists or when replacing object with map of reflected properties.
     * This might be implemented via "handlers" but now all logic is concentrated in single compact place...
     *
     * @param sw
     * @param jo
     * @param prefix
     * @param format
     * @param fakeObject - used to indicate the object is fake and represents currently stacked object (e.g. List for Array or Map for reflection)
     */
    public static void toFormattedString(StructureWriter sw, Object jo, String prefix, StructureFormat format, boolean fakeObject) throws IOException {
        //System.out.println("toFormattedString: " + format.getPath());
        if (jo == null) {
            sw.write(STRUCTURE_FORMAT_ELEMENT.Null, "null");
            return;
        }
        if (format.getStack().contains(jo) || format.getObjectsRegistry().isProcessedObject(jo) && format.getObjectsRegistry().isReferencable(jo)) {
            sw.write(STRUCTURE_FORMAT_ELEMENT.crossReferenceValue, format.getCrossReference(jo));
            return;
        }
        if (!fakeObject) {
            format.getStack().push(jo);
            format.getObjectsRegistry().registerProcessedObject(format.getPath().toString(), jo);
        }
        if (jo instanceof Map) {
            sw.write(STRUCTURE_FORMAT_ELEMENT.startMap, format.startMap((Map) jo));
            Map map = (Map) jo;
            // wrap map if needed
            map = ReflectiveBuilder.wrapMap(map);
            Collection keys = format.filteredKeys(map);


            // proceed with Map of string-object pairs
            int idx = 0;
            int maxIdx = keys.size();
            for (Object key : keys) {
                idx++;
                sw.write(STRUCTURE_FORMAT_ELEMENT.newLine, format.getNewLine());
                sw.write(STRUCTURE_FORMAT_ELEMENT.prefix, prefix);
                sw.write(STRUCTURE_FORMAT_ELEMENT.prefix, format.getIndentString());
                sw.write(STRUCTURE_FORMAT_ELEMENT.startElement, format.startElement(key.toString(), map.get(key)));
                sw.write(STRUCTURE_FORMAT_ELEMENT.valueSeparator, format.getValueSeparator());
                format.getPath().add(key.toString());
                toFormattedString(sw, map.get(key), prefix + format.getIndentString(), format, false);
                format.getObjectsRegistry().registerProcessedObject(format.getPath().toString(), map.get(key));
                sw.write(STRUCTURE_FORMAT_ELEMENT.endElement, format.endElement(key.toString(), map.get(key)));
                format.getPath().remove();
                if (idx < maxIdx) {
                    sw.write(STRUCTURE_FORMAT_ELEMENT.objectSeparator, format.getObjectSeparator());
                }
                if (idx == maxIdx) {
                    sw.write(STRUCTURE_FORMAT_ELEMENT.newLine, format.getNewLine());
                }
            }

            if (!keys.isEmpty()) {
                sw.write(STRUCTURE_FORMAT_ELEMENT.prefix, prefix);
            }
            sw.write(STRUCTURE_FORMAT_ELEMENT.endMap, format.endMap((Map) jo));
        } else if (jo instanceof List) {
            if (((List) jo).isEmpty()) {
                //sw.write(prefix);
                sw.write(STRUCTURE_FORMAT_ELEMENT.startList, format.startList((List) jo));
                sw.write(STRUCTURE_FORMAT_ELEMENT.endList, format.endList((List) jo));
            } else {
                sw.write(STRUCTURE_FORMAT_ELEMENT.startList, format.startList((List) jo));
                sw.write(STRUCTURE_FORMAT_ELEMENT.newLine, format.getNewLine());
                int idx = 0;
                int maxIdx = ((List) jo).size();
                for (Object sjo : ((List) jo)) {
                    //String pName = format.getPath().remove();
                    //format.getPath().add(pName, idx);
                    format.getPath().setIndex(idx, true);
                    idx++;
                    sw.write(STRUCTURE_FORMAT_ELEMENT.prefix, prefix);
                    sw.write(STRUCTURE_FORMAT_ELEMENT.prefix, format.getIndentString());
                    toFormattedString(sw, sjo, prefix + format.getIndentString(), format, false);
                    format.getObjectsRegistry().registerProcessedObject(format.getPath().toString(), sjo);
                    if (idx < maxIdx) {
                        sw.write(STRUCTURE_FORMAT_ELEMENT.objectSeparator, format.getObjectSeparator());
                    }
                    if (idx < maxIdx) {
                        sw.write(STRUCTURE_FORMAT_ELEMENT.newLine, format.getNewLine());
                    }
                }
                sw.write(STRUCTURE_FORMAT_ELEMENT.newLine, format.getNewLine());
                sw.write(STRUCTURE_FORMAT_ELEMENT.prefix, prefix);
                sw.write(STRUCTURE_FORMAT_ELEMENT.endList, format.endList((List) jo));
            }
        } else if (jo.getClass().isArray()) {
            toFormattedString(sw, Arrays.asList((Object[]) jo), prefix, format, true);
        } else if (format.testIsString(jo)) {
            sw.write(STRUCTURE_FORMAT_ELEMENT.startValue, format.startValue(jo));
            sw.write(STRUCTURE_FORMAT_ELEMENT.scalarValue, format.getFormattedScalarValue(jo));
            sw.write(STRUCTURE_FORMAT_ELEMENT.endValue, format.endValue(jo));
        } else {
            sw.write(STRUCTURE_FORMAT_ELEMENT.startValue, format.startValue(jo));
            if (jo.getClass().getSimpleName().equalsIgnoreCase("Null")) {
                // instanceof org.codehaus.groovy.grails.web.json.JSONObject.Null or any other null-indicator...) {
                toFormattedString(sw, null, prefix + format.getIndentString(), format, false);
            } else {
                Pair<PRIMITIVE_TYPE, String> value = format.getFormattedScalarValue(jo);
                if (value.getA() != PRIMITIVE_TYPE.none) {
                    sw.write(STRUCTURE_FORMAT_ELEMENT.scalarValue, value);
                } else {
                    Map obj2map = format.getReflectiveBuilder().getReflectable(jo);
                    if (!obj2map.isEmpty()) {
                        //toFormattedString(sw, obj2map, prefix + format.getIndentString(), format, true);
                        toFormattedString(sw, obj2map, prefix, format, true);
                    } else {
                        sw.write(STRUCTURE_FORMAT_ELEMENT.stringValue, format.getFormattedScalarValue(jo.toString()));
                    }
                }
            }
            sw.write(STRUCTURE_FORMAT_ELEMENT.endValue, format.endValue(jo));
        }
        if (!fakeObject) {
            format.getStack().pop();
        }
    }

    /**
     * Used to wrap provided object in map with specified name as key when applicable.
     * 
     * @param object
     * @param name
     * @return 
     */
    public static Object ensureObjectName(Object object, String name) {
        if (object == null || name == null || name.trim().length() == 0) {
            return object;
        } else if (object instanceof List) {
            return object;
        } else if (object instanceof String) {
            return object;
        } else {
            Map wrapper = new HashMap();
            wrapper.put(name, object);
            return wrapper;
        }
    }
}
