/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml;

import com.ssg.tools.jsonxml.Comparer.COMPARE_STATUS;
import com.ssg.tools.jsonxml.Comparer.ComparatorContext;
import com.ssg.tools.jsonxml.Comparer.ComparatorPair;
import com.ssg.tools.jsonxml.impl.DumpFormat;
import com.ssg.tools.jsonxml.impl.JSONFormat;
import com.ssg.tools.jsonxml.impl.JSONParser;
import com.ssg.tools.jsonxml.impl.StructureFormat;
import com.ssg.tools.jsonxml.impl.XMLFormat;
import com.ssg.tools.jsonxml.impl.XMLParser;
import com.ssg.tools.jsonxml.utils.Formats;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

/**
 *
 * @author ssg
 */
public class diff {

    public static void help() {
        System.out.println("Comparator tool: reads 2 JSON and or XML texts and writes to STDOUT comparison results.");
        System.out.println(" diff [-format=JSON|XML|Dump] [-indent=true|false] [-locale=locale] [-depth=number] [-delta=true|false] [-help] <1st input file or URL> <2nd input file or URL>");
        System.out.println("   -format - defines output format. Default is JSON.");
        System.out.println("   -indent - sets/clears output indentation. Default is true.");
        System.out.println("   -locale - Name of locale to use when parsing input.");
        System.out.println("   -depth  - Maximal nesint depth when comparing. Default is 100.");
        System.out.println("   -delta  - Shows only mismatching nodes info if true. Default is false.");
        System.out.println("   1st file  - Path or URL for first compared file.  In oputput it corresponds to A.");
        System.out.println("   2nd file  - Path or URL for second compared file. In oputput it corresponds to B.");
    }

    public static void main(String[] args) throws Exception {
//        LogManager lm = LogManager.getLogManager();
//        lm.readConfiguration(Thread.currentThread().getContextClassLoader().getResourceAsStream("logger.test.properties"));
//        Logger log = Logger.getLogger(converter.class.getName());
        boolean indent = true;
        String format = "JSON";
        Locale locale = Locale.getDefault();
        String path1 = null;
        String path2 = null;
        int depth = 100;
        boolean delta = false;

        if (args.length == 0) {
            help();
            return;
        }
        for (String arg : args) {
            if (arg.startsWith("-h") || arg.startsWith("-H") || arg.startsWith("-?")) {
                help();
                return;
            } else if (arg.startsWith("-format=")) {
                format = arg.substring("-format=".length());
                if (format.equalsIgnoreCase("json")) {
                    format = "JSON";
                } else if (format.equalsIgnoreCase("XML")) {
                    format = "XML";
                } else if (format.equals("Dump")) {
                    format = "Dump";
                } else {
                    help();
                    System.out.println("  wrong parameter: " + arg + ". Error: unknown output format.");
                }
            } else if (arg.startsWith("-indent=")) {
                try {
                    indent = Boolean.parseBoolean(arg.substring("-indent=".length()));
                } catch (Throwable th) {
                    help();
                    System.out.println("  wrong parameter: " + arg + ". Error: " + th);
                }
            } else if (arg.startsWith("-locale=")) {
                try {
                    locale = new Locale(arg.substring("-locale=".length()));
                } catch (Throwable th) {
                    help();
                    System.out.println("  wrong parameter: " + arg + ". Error: " + th);
                }
            } else if (arg.startsWith("-depth=")) {
                try {
                    depth = Integer.parseInt(arg.substring("-depth=".length()));
                } catch (NumberFormatException nfex) {
                    help();
                    System.out.println("  wrong parameter: " + arg + ". Error: " + nfex);
                }
            } else if (arg.startsWith("-delta=")) {
                try {
                    delta = Boolean.parseBoolean(arg.substring("-delta=".length()));
                } catch (Throwable th) {
                    help();
                    System.out.println("  wrong parameter: " + arg + ". Error: " + th);
                }
            } else if (arg.startsWith("-")) {
                help();
                System.out.println("  unknown parameter: " + arg + ".");
            } else {
                if (path1 == null) {
                    path1 = arg;
                } else if (path2 == null) {
                    path2 = arg;
                } else {
                    help();
                    System.out.println("  wrong parameter: " + arg + ". Only 2 files (URLs) are accepted. This is 3rd.");
                    System.out.println("    1st file (URL): " + path1);
                    System.out.println("    2nd file (URL): " + path2);
                }
            }
        }

        if (path1 == null || path2 == null) {
            help();
            System.out.println("  ERROR: both input file name(s) or URL(s) are required.");
            System.out.println("    1st file (URL): " + path1);
            System.out.println("    2nd file (URL): " + path2);
            return;
        }

        Formats formats = new Formats(locale, null);
        Object obj1 = parsePath(path1, formats, locale);
        Object obj2 = parsePath(path2, formats, locale);

        if (obj1 == null || obj2 == null) {
            System.out.println("  ERROR: either of both input files could not be parsed:");
            System.out.println("    1st file (URL): " + path1 + " parsed=" + (obj1 != null));
            System.out.println("    2nd file (URL): " + path2 + " parsed=" + (obj2 != null));
            return;
        }

        Writer writer = new OutputStreamWriter(System.out);

        StructureFormat sf = null;
        if (format.equals("JSON")) {
            sf = new JSONFormat(indent);
        } else if (format.equals("XML")) {
            sf = new XMLFormat(indent);
        } else if (format.equals("Dump")) {
            sf = new DumpFormat(indent);
        }

        ComparatorContext context = new ComparatorContext();
        Comparer diff = new Comparer();
        ComparatorPair result = diff.compare(context, obj1, obj2, depth);

        if (delta) {
            removeMatching(result);
        }

        Formatter.toFormattedString(writer, result, "", sf, false);
        writer.close();
    }

    public static Object parsePath(String path, Formats formats, Locale locale) {
        Object obj = null;
        Throwable xmlErr = null;
        Throwable jsonErr = null;
        try {
            Reader r = new InputStreamReader((path.contains("//")) ? (new URL(path)).openStream() : new FileInputStream(new File(path)));
            obj = (new XMLParser()).parse(r, formats);
        } catch (Throwable th) {
            xmlErr = th;
        }
        if (obj == null) {
            try {
                Reader r = new InputStreamReader((path.contains("//")) ? (new URL(path)).openStream() : new FileInputStream(new File(path)));
                obj = (new JSONParser()).parse(r, new Formats(locale, null));
            } catch (Throwable th) {
                jsonErr = th;
            }
        }
        if (obj == null) {
            help();
            System.out.println("  ERROR: failed to parse as XML : " + xmlErr);
            System.out.println("  ERROR: failed to parse as JSON: " + jsonErr);
        }
        return obj;

    }

    public static void removeMatching(ComparatorPair cp) {
        List<ComparatorPair> nested = cp.getNested();
        if (nested == null || nested.isEmpty()) {
            return;
        }
        List<ComparatorPair> toRemove = new ArrayList<ComparatorPair>();
        for (ComparatorPair scp : nested) {
            if (scp.getStatus() == COMPARE_STATUS.match) {
                toRemove.add(scp);
            } else {
                removeMatching(scp);
            }
        }
        for (ComparatorPair rcp : toRemove) {
            nested.remove(rcp);
        }
    }
}
