/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.impl;

import com.ssg.tools.jsonxml.utils.Formats;
import com.ssg.tools.jsonxml.utils.PRIMITIVE_TYPE;
import com.ssg.tools.jsonxml.utils.Utilities;
import com.ssg.tools.jsonxml.utils.Utilities.Pair;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;

/**
 * Provides values for proper formatting of JSON objects.
 *
 * @author ssg
 */
public class JSONFormat extends StructureFormat {

    private boolean _strict = true;

    public JSONFormat() {
        init(false);
    }

    public JSONFormat(boolean indented) {
        init(indented);
    }

    public JSONFormat(boolean indented, boolean strict) {
        init(indented);
    }

    @Override
    protected void init(boolean indented) {
        setPrefix("");
        setNewLine((indented) ? "\n" : "");
        setValueSeparator((indented) ? ": " : ":");
        setObjectSeparator(",");
        setIndentString((indented) ? "    " : "");
        setQuoteName("\"");
        setQuoteValue("\"");
        getReplacements().clear();
        getReplacements().put("\\", "\\\\");
        getReplacements().put("/", "\\/");
        getReplacements().put("\n", "\\n");
        getReplacements().put("\r", "\\r");
        getReplacements().put("\t", "\\t");
        enforceStrictSyntax();
    }

    public void enforceStrictSyntax() {
        if (_strict) {
            setFormats(new Formats(Locale.US, getFormats().getTimeZone()));
            for (int i = 0; i < getFormats().getNumbers().size(); i++) {
                getFormats().getNumbers().get(i).setGroupingUsed(false);
            }
        }
    }

    /**
     * Performs character replacements in string values to match JSON specification.
     *
     * @param value
     * @param format
     * @return
     */
    @Override
    public String encodeStringValue(String value) {
        if (value == null) {
            return null;
        }
        value = super.encodeStringValue(value);
        // ensures quote used for value wrapping is not broken by value itself...
        value = value.replace(getQuoteValue(), "\\" + getQuoteValue());
        return value;
    }

    @Override
    public String decodeStringValue(String value) {
        value = value.replace("\\" + getQuoteValue(), getQuoteValue());
        return super.decodeStringValue(value);
    }

    @Override
    public Pair<PRIMITIVE_TYPE, String> getFormattedScalarValue(Object object) {
        Pair<PRIMITIVE_TYPE, String> result = super.getFormattedScalarValue(object);
        String s = result.getB();
        if (s != null
                && ((s.contains(getObjectSeparator())
                && Utilities.isNumericClass(object.getClass()))
                || object instanceof Date
                || result.getA() == PRIMITIVE_TYPE.enumeration)) {
            //System.out.println("getFormattedScalarValue: "+object+" ["+object.getClass()+", "+getFormats().isNumericClass(object.getClass()) +"] -> "+s);
            s = getQuoteValue() + s + getQuoteValue();
            result.setB(s);
        }
        return result;
    }

    @Override
    public String startMap(Map object) {
        return "{";
    }

    @Override
    public String endMap(Map object) {
        return "}";
    }

    @Override
    public String startList(List object) {
        return "[";
    }

    @Override
    public String endList(List object) {
        return "]";
    }

    @Override
    public String startValue(Object object) {
        return "";
    }

    @Override
    public String endValue(Object object) {
        return "";
    }

    /**
     * Used to pre-process element name when needed.
     * @param name
     * @return
     */
    public String elementName(String name) {
        return name;
    }
}
