/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.impl;

import com.ssg.tools.jsonxml.Parser;
import com.ssg.tools.jsonxml.utils.Formats;
import com.ssg.tools.jsonxml.utils.StructureErrorInfo;
import com.ssg.tools.jsonxml.utils.StructureProcessingException;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Logger;

/**
 *
 * @author ssg
 */
public class JSONParser extends Parser {

    static Logger log = Logger.getLogger(Parser.class.getName());

    public static class JSONParserContext extends ParserContext {

        public static Object OBJECT_DELIMITER_TAG = new Object();
        public static final char OPEN_LIST = '[';
        public static final char OPEN_OBJECT = '{';
        public static final char CLOSE_LIST = ']';
        public static final char CLOSE_OBJECT = '}';
        public static final char OBJECTS_DELIMITER = ',';
        public static final char VALUE_DELIMITER = ':';
        private JSONFormat _jsonFormat = new JSONFormat();
        // parsing position. Line/Position are 1-based. GlobalPosition is 0-based stream posiotion.
        private int _line;
        private long _position;
        private long _globalPosition;
        private boolean _eof = false;

        public JSONParserContext() {
        }

        public JSONParserContext(ParserContext context) {
            if (context != null) {
                setStack(context.getStack());
                setPath(context.getPath());
                setFormats(context.getFormats());
            }
        }

        public JSONParserContext(Locale locale) {
            super(locale);
        }

        /**
         * @return the _jsonFormat
         */
        public JSONFormat getJsonFormat() {
            return _jsonFormat;
        }

        /**
         * @param jsonFormat the _jsonFormat to set
         */
        public void setJsonFormat(JSONFormat jsonFormat) {
            this._jsonFormat = jsonFormat;
        }

        @Override
        public Formats getFormats() {
            if (_jsonFormat != null) {
                return _jsonFormat.getFormats();
            } else {
                return super.getFormats();
            }
        }

        @Override
        public void setFormats(Formats formats) {
            if (_jsonFormat != null) {
                _jsonFormat.setFormats(formats);
                _jsonFormat.enforceStrictSyntax();
            } else {
                super.setFormats(formats);
            }
        }

        /**
         * @return the _line
         */
        public int getLine() {
            return _line;
        }

        /**
         * @param line the _line to set
         */
        public void setLine(int line) {
            this._line = line;
        }

        /**
         * @return the _position
         */
        public long getPosition() {
            return _position;
        }

        /**
         * @param position the _position to set
         */
        public void setPosition(long position) {
            this._position = position;
        }

        /**
         * @return the _globalPosition
         */
        public long getGlobalPosition() {
            return _globalPosition;
        }

        /**
         * @param globalPosition the _globalPosition to set
         */
        public void setGlobalPosition(long globalPosition) {
            this._globalPosition = globalPosition;
        }

        /**
         * @return the _eof
         */
        public boolean isEOF() {
            return _eof;
        }

        /**
         * @param eof the _eof to set
         */
        public void setEOF(boolean eof) {
            this._eof = eof;
        }
    }

    public Object postProcessObject(JSONParserContext context, Object object) {
        //System.out.println("postProcessObject: "+context.getPath()+"; "+((object!=null) ? object.getClass().getSimpleName() : "null"));
        Object primitive = context.getFormats().toPrimitiveValue(object);
        if (primitive != null) {
            return primitive;
        } else if (object instanceof String) {
            return context.getJsonFormat().decodeStringValue((String) object);
        } else {
            return object;
        }
    }

    public Object parse(Reader r) throws IOException {
        return parse(r, (Formats) null);
    }

    public Object parse(Reader r, Formats formats) throws IOException {
        JSONParserContext context = new JSONParserContext();
        if (formats != null) {
            context.setFormats(formats);
        }
        return parse(r, context);
    }

    public Object parse(Reader r, ParserContext jsonContext) throws IOException {
        if (!r.markSupported()) {
            r = new BufferedReader(r);
        }

        final JSONParserContext context = (jsonContext != null)
                ? ((jsonContext instanceof JSONParserContext))
                ? (JSONParserContext) jsonContext
                : new JSONParserContext(jsonContext)
                : new JSONParserContext();

        // "inject" line/position tracker for diagnosis...
        r = new BufferedReader(r) {

            int buf_limit = 0;
            int line = 1;
            long pos = 1;
            long globalPos = 0;
            int old_line;
            long old_pos;
            int last_ch = 0;

            @Override
            public void mark(int readAheadLimit) throws IOException {
                super.mark(readAheadLimit);
                buf_limit = readAheadLimit;
                old_line = line;
                old_pos = pos;
            }

            @Override
            public int read() throws IOException {
                int ch = super.read();
                if (ch != READER_EOF) {
                    globalPos++;
                    pos++;
                    if (ch == '\r' || ch == '\n') {
                        pos = 1;
                        if (last_ch != ch && (last_ch == '\n' || last_ch == '\r')) {
                            line++;
                        }
                    }
                    last_ch = ch;
                }
                copyToContext();
                return ch;
            }

            @Override
            public void reset() throws IOException {
                super.reset();
                globalPos -= buf_limit;
                line = old_line;
                pos = old_pos;
                copyToContext();
            }

            void copyToContext() {
                context.setLine(line);
                context.setPosition(pos);
                context.setGlobalPosition(globalPos);
            }
        };

        return parseJSON(r, false, context);
    }

    /**
     * special notes on context._path usage:
     *   - added for maps.
     *   - for created list existing name is enforced to be array.
     * @param r
     * @param asName
     * @param context
     * @return
     * @throws IOException
     */
    Object parseJSON(Reader r, boolean asName, JSONParserContext context) throws IOException {
        StringBuilder sb = new StringBuilder();
        Object result = null;
        char ch = nextNonSpaceChar(r);
        switch (ch) {
            case READER_EOF:
                context.setEOF(true);
                break;
            case JSONParserContext.OPEN_LIST:
                result = new ArrayList();
                context.getStack().push(result);
                if (1 == 1) {
                    // ensure path is ARRAY now
                    String last = context.getPath().remove();
                    context.getPath().add(last, 0);
                }
                // iterate...
                if (result != null) {
                    Object value = parseJSON(r, false, context);
                    while (!result.equals(value) && !context.isEOF()) {
                        if (!JSONParserContext.OBJECT_DELIMITER_TAG.equals(value)) {
                            ((List) result).add(value);
                        }
                        String last = context.getPath().remove();
                        context.getPath().add(last, ((List) result).size());
                        value = parseJSON(r, false, context);
                    }
                    if (context.isEOF()) {
                        //System.out.println("parseJSON: ABNORMAL LIST END @"+context.getPath());
                        context.addError(new StructureErrorInfo(" ABNORMAL LIST END @" + context.getPath(), "" + context.CLOSE_LIST, "EOF"));
                    }
                }
                break;
            case JSONParserContext.CLOSE_LIST:
                if (context.getStack().peek() instanceof List) {
                    result = context.getStack().pop();
                    //context._path.remove();
                } else {
                    StructureErrorInfo error = new StructureErrorInfo(context.getLine(), (int) context.getPosition(), "'" + context.getJsonFormat().getObjectSeparator() + "' or '}'", "'" + ch + "'");
                    context.addError(error);
                    throw new StructureProcessingException("No open list to close.", null, error);
                }
                break;
            case JSONParserContext.OPEN_OBJECT:
                result = new LinkedHashMap();
                context.getStack().push(result);
                // iterate...
                if (result != null) {
                    Object name = parseJSON(r, true, context);
                    Object value = null;
                    if (!result.equals(name)) {
                        context.getPath().add(name.toString());
                        value = parseJSON(r, false, context);
                        context.getPath().remove();
                    }
                    while (name != null && !result.equals(name) && !context.isEOF()) {
                        if (!JSONParserContext.OBJECT_DELIMITER_TAG.equals(name)) {
                            ((Map) result).put(name, value);
                        }
                        name = parseJSON(r, true, context);
                        if (name != null && !result.equals(name) && !JSONParserContext.OBJECT_DELIMITER_TAG.equals(name)) {
                            context.getPath().add(name.toString());
                            value = parseJSON(r, false, context);
                            context.getPath().remove();
                        } else {
                            value = null;
                        }
                    }
                    if (context.isEOF()) {
                        //System.out.println("parseJSON: ABNORMAL OBJECT END @"+context.getPath());
                        context.addError(new StructureErrorInfo("ABNORMAL OBJECT END @" + context.getPath(), "" + context.CLOSE_OBJECT, "EOF"));
                    }
                }
                break;
            case JSONParserContext.CLOSE_OBJECT:
                if (context.getStack().peek() instanceof Map) {
                    result = context.getStack().pop();
                } else {
                    StructureErrorInfo error = new StructureErrorInfo(context.getLine(), (int) context.getPosition(), "'" + context.getJsonFormat().getObjectSeparator() + "' or ']'", "'" + ch + "'");
                    context.addError(error);
                    throw new StructureProcessingException("No open object to close.", null, error);
                }
                break;
            case JSONParserContext.OBJECTS_DELIMITER:
                result = JSONParserContext.OBJECT_DELIMITER_TAG;
                break;
            default:
                if (asName) {
                    String name = null;
                    if (ch == '\'' || ch == '"') {
                        name = readQuotedString(r, ch);
                        readQuotedString(r, JSONParserContext.VALUE_DELIMITER);
                    } else {
                        name = readQuotedString(r, JSONParserContext.VALUE_DELIMITER);
                    }
                    result = name;
                } else {
                    if (ch == '\'' || ch == '"') {
                        result = readQuotedString(r, ch);
                        result = ((String) result).replace("\\" + ch, "" + ch);
                    } else {
                        String s = ch + readNotString(r);
                        if (s == "null") {
                            return null;
                        } else {
                            Object primitive = context.getFormats().toPrimitiveValue(s);
                            if (primitive != null) {
                                result = primitive;
                            } else {
                                result = s;
                            }

                            if (1 == 0) {
                                try {
                                    result = Integer.parseInt(s);
                                } catch (Throwable th) {
                                    try {
                                        result = Double.parseDouble(s);
                                    } catch (Throwable th2) {
                                        try {
                                            result = Boolean.parseBoolean(s);
                                        } catch (Throwable th3) {
                                            result = s;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                break;
        }
        if (!asName) {
            result = postProcessObject(context, result);
        }
        context.getObjectsRegistry().registerProcessedObject(context.getPath().toString(), result);
        return result;
    }

    String readQuotedString(Reader r, char quoteChar) throws IOException {
        StringBuilder sb = new StringBuilder();
        char ch = readChar(r);
        // TODO: check for escaped quote char to avoid false string ending
        while (ch != READER_EOF && ch != quoteChar) {
            sb.append(ch);
            ch = readChar(r);
        }
        return sb.toString();
    }

    String readNotString(Reader r) throws IOException {
        StringBuilder sb = new StringBuilder();
        r.mark(1);
        char ch = readChar(r);
        while (ch != READER_EOF && !isWhiteSpace(ch) && ch != JSONParserContext.CLOSE_OBJECT && ch != JSONParserContext.CLOSE_LIST && ch != JSONParserContext.OBJECTS_DELIMITER) {
            sb.append(ch);
            r.mark(1);
            ch = readChar(r);
        }
        if (ch == JSONParserContext.CLOSE_OBJECT || ch == JSONParserContext.CLOSE_LIST || ch == JSONParserContext.OBJECTS_DELIMITER) {
            r.reset();
        }
        return sb.toString();
    }
}
