/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.impl;

import com.ssg.tools.jsonxml.utils.PRIMITIVE_TYPE;
import com.ssg.tools.jsonxml.utils.StructureProcessingContext;
import com.ssg.tools.jsonxml.utils.Utilities.Pair;
import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

/**
 * Default structure format.
 * Provides values/methods for formatting.
 * Provides entry points for controlling reflection-based formatting and implements basic one.
 * 
 * 
 *
 * @author ssg
 */
public class StructureFormat extends StructureProcessingContext {

    private String prefix = "";
    private String newLine = "\n";
    private String valueSeparator = "=";
    private String objectSeparator = ",";
    private String indentString = "  ";
    private String quoteName = "";
    private String quoteValue = "";
    private List<String> skip = new ArrayList<String>();
    private Map<String, String> replacements = new LinkedHashMap<String, String>() {

        {
            put("\n", "\\n");
            put("\r", "\\r");
            put("\t", "\\t");
        }
    };

    public StructureFormat() {
        init(false);
    }

    public StructureFormat(boolean indented) {
        init(indented);
    }

    protected void init(boolean indented) {
        if (!indented) {
            setPrefix("");
            setNewLine("");
            setIndentString("  ");
        }
    }

    /**
     * Returns properly formatted JSON value: number, true, false or string.
     * If not "formattable" (e.g. object or null) null is returned.
     * 
     * @param object
     * @return 
     */
    public Pair<PRIMITIVE_TYPE, String> getFormattedScalarValue(Object object) {
        Pair<PRIMITIVE_TYPE, String> result = getFormats().formatAndType(getPath(), object, -1);
        if (result.getA() == PRIMITIVE_TYPE.none && object != null) {
            if (object instanceof String || object.getClass().isEnum()) {
                result = new Pair<PRIMITIVE_TYPE, String>(PRIMITIVE_TYPE.string, getQuoteValue() + encodeStringValue(object.toString()) + getQuoteValue());
            }
        }
        return result;
    }

    /**
     * Provides filtering based on predefined set of keys in collection "skip".
     *
     * @param obj
     * @return
     */
    public Collection<String> filteredKeys(Map obj) {
        Collection<String> result = new ArrayList<String>(obj.size());
        if (getSkip().isEmpty()) {
            result.addAll(obj.keySet());
        } else {
            for (Object key : obj.keySet()) {
                if (key == null) {
                    continue;
                }
                boolean found = false;
                String s = key.toString();
                for (String expr : getSkip()) {
                    if (s.matches(expr)) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    result.add(s);
                }
            }
        }
        return result;
    }

    /**
     * Override to enforce custom objects to be formatted (rendered) as string.
     * By default interprets as String any File and any class having ".log4j." in qualified name.
     *
     * @param object
     * @return
     */
    public boolean testIsString(Object object) {
        if (object == null) {
            return false;
        }
        if (object instanceof String) {
            return true;
        } else {
            return object instanceof File || object.getClass().getName().contains(".log4j.");
        }
    }

    /**
     * Performs character replacements in string values to match JSON specification.
     *
     * @param value
     * @param format
     * @return
     */
    public String encodeStringValue(String value) {
        if (value == null) {
            return null;
        }
        if (getReplacements() != null) {
            for (String key : getReplacements().keySet()) {
                value = value.replace(key, getReplacements().get(key));
            }
        }
        return value;
    }

    /**
     * Performs reverse operation encodeStringValue.
     *
     * @param value
     * @param format
     * @return
     */
    public String decodeStringValue(String value) {
        if (value == null) {
            return null;
        }
        if (getReplacements() != null) {
            Stack<String> ks = new Stack<String>();
            for (String key : getReplacements().keySet()) {
                ks.push(key);
            }
            while (!ks.isEmpty()) {
                String key = ks.pop();
                value = value.replace(getReplacements().get(key), key);
            }
        }
        return value;
    }

    /**
     * Returns string representing reference to existing object.
     * 
     * @param object
     * @return 
     */
    public String getCrossReference(Object object) {
        if (object != null && getObjectsRegistry().isReferencable(object) && getStack().contains(object)) {
            String s = getQuoteValue() + getObjectsRegistry().buildReferenceString(object.getClass(), this.getPath().getPathForLevel(getStack().indexOf(object)).toString()) + getQuoteValue();

            getObjectsRegistry().registerReferencedObject(getPath().getPathForLevel(getStack().indexOf(object)).toString(), object);
            return s;
        } else if (object != null && getObjectsRegistry().isReferencable(object) && getObjectsRegistry().isProcessedObject(object)) {
            String objectPath = getObjectsRegistry().getProcessedObjectPath(object);
            String s = getQuoteValue() + getObjectsRegistry().buildReferenceString(object.getClass(), objectPath) + getQuoteValue();

            getObjectsRegistry().registerReferencedObject(objectPath, object);
            return s;
        } else {
            return "";
        }
    }

    public String startMap(Map object) {
        return "{";
    }

    public String endMap(Map object) {
        return "}";
    }

    public String startList(List object) {
        return "[";
    }

    public String endList(List object) {
        return "]";
    }

    public String startElement(String name, Object object) {
        return getQuoteName() + name + getQuoteName();
    }

    public String endElement(String name, Object object) {
        return "";
    }

    public String startValue(Object object) {
        return "";
    }

    public String endValue(Object object) {
        return "";
    }

    /**
     * @return the prefix
     */
    public String getPrefix() {
        return prefix;
    }

    /**
     * @param prefix the prefix to set
     */
    public void setPrefix(String prefix) {
        this.prefix = prefix;
    }

    /**
     * @return the newLine
     */
    public String getNewLine() {
        return newLine;
    }

    /**
     * @param newLine the newLine to set
     */
    public void setNewLine(String newLine) {
        this.newLine = newLine;
    }

    /**
     * @return the valueSeparator
     */
    public String getValueSeparator() {
        return valueSeparator;
    }

    /**
     * @param valueSeparator the valueSeparator to set
     */
    public void setValueSeparator(String valueSeparator) {
        this.valueSeparator = valueSeparator;
    }

    /**
     * @return the objectSeparator
     */
    public String getObjectSeparator() {
        return objectSeparator;
    }

    /**
     * @param objectSeparator the objectSeparator to set
     */
    public void setObjectSeparator(String objectSeparator) {
        this.objectSeparator = objectSeparator;
    }

    /**
     * @return the indentString
     */
    public String getIndentString() {
        return indentString;
    }

    /**
     * @param indentString the indentString to set
     */
    public void setIndentString(String indentString) {
        this.indentString = indentString;
    }

    /**
     * @return the quoteName
     */
    public String getQuoteName() {
        return quoteName;
    }

    /**
     * @param quoteName the quoteName to set
     */
    public void setQuoteName(String quoteName) {
        this.quoteName = quoteName;
    }

    /**
     * @return the quoteValue
     */
    public String getQuoteValue() {
        return quoteValue;
    }

    /**
     * @param quoteValue the quoteValue to set
     */
    public void setQuoteValue(String quoteValue) {
        this.quoteValue = quoteValue;
    }

    /**
     * @return the skip
     */
    public List<String> getSkip() {
        return skip;
    }

    /**
     * @param skip the skip to set
     */
    public void setSkip(List<String> skip) {
        this.skip = skip;
    }

    /**
     * @return the replacements
     */
    public Map<String, String> getReplacements() {
        return replacements;
    }

    /**
     * @param replacements the replacements to set
     */
    public void setReplacements(Map<String, String> replacements) {
        this.replacements = replacements;
    }
}
