/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.utils;

import com.ssg.tools.jsonxml.utils.Utilities.Pair;
import java.lang.reflect.Constructor;
import java.math.BigDecimal;
import java.text.DateFormat;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;

/**
 * Collection of formatters for handling objects.
 * Use getters to view/remove formats, BUT use add* methods to add new format.
 *
 * Used in parsers to try get instance of primitive type with toPrimitiveType(...)
 *
 * @author ssg
 */
public class Formats {

    /**
     * This is number of digits in number from which BiDecimal instantiation will be tried,.
     * Otherwise double value is used.
     */
    private static byte BIGDECIMAL_DIGITS_FROM = 20;
    private static Constructor _bigDecimalConstructor;
    private static DecimalFormatSymbols _decimalSymbolsUS = DecimalFormatSymbols.getInstance(Locale.US);
    private Locale _locale = Locale.getDefault();
    private TimeZone _timeZone = TimeZone.getDefault();
    private DecimalFormatSymbols decimalSymbols;
    private List<DateFormat> _dates = new ArrayList<DateFormat>();
    private List<DateFormat> _times = new ArrayList<DateFormat>();
    private List<DateFormat> _dateTimes = new ArrayList<DateFormat>();
    private List<NumberFormat> _numbers = new ArrayList<NumberFormat>();

    static {
        try {
            _bigDecimalConstructor = BigDecimal.class.getConstructor(String.class);
        } catch (Throwable th) {
        }

    }

    public Formats() {
        init();
    }

    public Formats(Locale locale, TimeZone timeZone) {
        if (locale != null) {
            _locale = locale;
        }
        if (timeZone == null) {
            _timeZone = timeZone;
        }
        init();
    }

    private void init() {
        NumberFormat nf = NumberFormat.getNumberInstance(_locale);
        _numbers.add(nf);

        DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT, _locale);
        df.setTimeZone(_timeZone);
        _dates.add(df);

        df = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT, _locale);
        df.setTimeZone(_timeZone);
        _dateTimes.add(df);

        df = DateFormat.getTimeInstance(DateFormat.SHORT, _locale);
        df.setTimeZone(_timeZone);
        _times.add(df);

        decimalSymbols = DecimalFormatSymbols.getInstance(_locale);
    }

    public void addDateFormat(String pattern) {
        DateFormat df = new SimpleDateFormat(pattern, getLocale());
        df.setTimeZone(getTimeZone());
        if (!_dates.contains(df)) {
            getDates().add(df);
        }
    }

    public void addTimeFormat(String pattern) {
        DateFormat df = new SimpleDateFormat(pattern, getLocale());
        df.setTimeZone(getTimeZone());
        if (!_times.contains(df)) {
            getTimes().add(df);
        }
    }

    public void addDateTimeFormat(String pattern) {
        DateFormat df = new SimpleDateFormat(pattern, getLocale());
        df.setTimeZone(getTimeZone());
        if (!_dateTimes.contains(df)) {
            getDateTimes().add(df);
        }
    }

    Boolean asBoolean(Object value) {
        try {
            if (value == null) {
                return null;
            } else if (value.toString().equals("true")) {
                return true;
            } else if (value.toString().equals("false")) {
                return false;
            } else {
                return null;
            }
        } catch (Throwable th) {
            return null;
        }
    }

    Date asDate(Object value) {
        for (DateFormat df : getDates()) {
            try {
                return df.parse(value.toString());
            } catch (Throwable th) {
                // ignore - we are searching for any matching format.
            }
        }
        return null;
    }

    Date asTime(Object value) {
        for (DateFormat df : getTimes()) {
            try {
                return df.parse(value.toString());
            } catch (Throwable th) {
                // ignore - we are searching for any matching format.
            }
        }
        return null;
    }

    Date asDateTime(Object value) {
        for (DateFormat df : getDateTimes()) {
            try {
                return df.parse(value.toString());
            } catch (Throwable th) {
                // ignore - we are searching for any matching format.
            }
        }
        return null;
    }

    Number asNumber(Object value) {
        for (NumberFormat nf : getNumbers()) {
            try {
                String s = value.toString().trim();
                ParsePosition pos = new ParsePosition(0);
                Number n = nf.parse(value.toString(), pos);
                if (pos.getErrorIndex() != -1) {
                    return null;
                } else if (pos.getIndex() < (s.length() - 1)) {
                    if (s.indexOf(decimalSymbols.getExponentSeparator()) != -1) {
                        return Double.parseDouble(toUSNumber(s));
                    } else {
                        return null;
                    }
                } else {
                    if (n instanceof Long && s.indexOf(decimalSymbols.getDecimalSeparator()) != -1) {
                        n = n.doubleValue();
                    } else if (n instanceof Double) {
                        if (s.indexOf(decimalSymbols.getExponentSeparator()) == -1 && s.length() > BIGDECIMAL_DIGITS_FROM) {
                            try {
                                //BigDecimal d = new BigDecimal(s);
                                //return d;
                                return (Number) _bigDecimalConstructor.newInstance(toUSNumber(s));
                            } catch (Throwable th) {
                                return n;
                            }
                        }
                    }
                    return n;
                }
            } catch (Throwable th) {
            }
        }
        return null;
    }

    public Number asNumber(Object value, Class clazz) {
        Number obj = asNumber(value);
        if (obj == null || clazz == null) {
            return obj;
        }
        if (clazz.equals(Byte.class) || clazz.equals(byte.class)) {
            return obj.byteValue();
        } else if (clazz.equals(Short.class) || clazz.equals(short.class)) {
            return obj.shortValue();
        } else if (clazz.equals(Integer.class) || clazz.equals(int.class)) {
            return obj.intValue();
        } else if (clazz.equals(Long.class) || clazz.equals(long.class)) {
            return obj.longValue();
        } else if (clazz.equals(Float.class) || clazz.equals(float.class)) {
            return obj.floatValue();
        } else if (clazz.equals(Double.class) || clazz.equals(double.class)) {
            return obj.doubleValue();
        } else if (clazz.equals(BigDecimal.class)) {
            if (obj instanceof BigDecimal) {
                return obj;
            } else {
                return new BigDecimal(obj.toString());
            }
        }
        return obj;
    }

    private String toUSNumber(String s) {
        if (s == null) {
            return s;
        }
        return s.replace("" + decimalSymbols.getGroupingSeparator(), "").
                replace(decimalSymbols.getExponentSeparator(), _decimalSymbolsUS.getExponentSeparator()).
                replace(decimalSymbols.getDecimalSeparator(), _decimalSymbolsUS.getDecimalSeparator());
    }

    public Object toPrimitiveValue(Object object) {
        Pair<PRIMITIVE_TYPE, Object> result = toPrimitiveTypeValue(object);
        return result.getB();
    }

    public Pair<PRIMITIVE_TYPE, Object> toPrimitiveTypeValue(Object object) {
        // used to convert object (string?) to nearest primitive type or null
        Pair<PRIMITIVE_TYPE, Object> result = new Pair<PRIMITIVE_TYPE, Object>(PRIMITIVE_TYPE.none, null);
        if (object == null) {
            return result;
        }
        result.setB(asDateTime(object));
        if (result.getB() != null) {
            result.setA(PRIMITIVE_TYPE.datetime);
        } else {
            result.setB(asTime(object));
            if (result.getB() != null) {
                result.setA(PRIMITIVE_TYPE.time);
            } else {
                result.setB(asDate(object));
                if (result.getB() != null) {
                    result.setA(PRIMITIVE_TYPE.date);
                } else {
                    result.setB(asNumber(object));
                    if (result.getB() != null) {
                        result.setA(PRIMITIVE_TYPE.number);
                    } else {
                        result.setB(asBoolean(object));
                        if (result.getB() != null) {
                            result.setA(PRIMITIVE_TYPE.bool);
                        } else if (object.getClass().isEnum()) {
                            result.setB(object.toString());
                            result.setA(PRIMITIVE_TYPE.enumeration);
                            System.out.println("formatted " + object + " (" + object.getClass().getName() + ") as enumeration.");
                        }
                    }
                }
            }
        }
        return result;
    }

    public Calendar getCalendar() {
        return Calendar.getInstance(_timeZone, _locale);
    }

    public DateFormat getDateFormat(String mask) {
        DateFormat df = new SimpleDateFormat(mask, _locale);
        df.setTimeZone(_timeZone);
        return df;
    }

    /**
     * Formats object using provided index or 0 if index is invalid
     * @param object
     * @param index
     * @return 
     */
    public String format(Path path, Object object, int index) {
        //System.out.println("Formats.format: "+path+" ["+((object!=null) ? object.getClass().getSimpleName() : "null")+", "+index+"]: "+object);
        Pair<PRIMITIVE_TYPE, String> result = formatAndType(path, object, index);
        return result.getB();
    }

    public Pair<PRIMITIVE_TYPE, String> formatAndType(Path path, Object object, int index) {
        Pair<PRIMITIVE_TYPE, String> result = new Pair<PRIMITIVE_TYPE, String>(PRIMITIVE_TYPE.none, null);

        if (index < 0) {
            index = 0;
        }

        if (object instanceof Byte || object instanceof Integer || object instanceof Long || object instanceof Float || object instanceof Double || object instanceof BigDecimal) {
            if (index >= _numbers.size()) {
                index = 0;
            }
            result.setB(_numbers.get(0).format(object));
            result.setA(PRIMITIVE_TYPE.number);
        } else if (object instanceof Boolean) {
            result.setB(object.toString());
            result.setA(PRIMITIVE_TYPE.bool);
        } else if (object instanceof Date) {
            Date d = (Date) object;
            String mask = "yyyy-DD-MM HH:mm:sss";
            DateFormat df = getDateFormat(mask);
            mask = df.format(d);

            if (mask.startsWith("1970-01-01")) {
                if (index >= _times.size()) {
                    index = 0;
                }

                result.setB(_times.get(index).format((Date) object));
                result.setA(PRIMITIVE_TYPE.time);
            } else if (mask.endsWith("00:00:000")) {
                if (index >= _dates.size()) {
                    index = 0;
                }
                result.setB(_dates.get(index).format((Date) object));
                result.setA(PRIMITIVE_TYPE.date);
            } else {
                if (index >= _dateTimes.size()) {
                    index = 0;
                }
                result.setB(_dateTimes.get(index).format((Date) object));
                result.setA(PRIMITIVE_TYPE.datetime);
            }
        } else if (object.getClass().isEnum()) {
            result.setB(object.toString());
            result.setA(PRIMITIVE_TYPE.enumeration);
        }
        return result;
    }

    /**
     * @return the _locale
     */
    public Locale getLocale() {
        return _locale;
    }

    /**
     * @return the _timeZone
     */
    public TimeZone getTimeZone() {
        return _timeZone;
    }

    /**
     * @return the _dates
     */
    public List<DateFormat> getDates() {
        return _dates;
    }

    /**
     * @return the _times
     */
    public List<DateFormat> getTimes() {
        return _times;
    }

    /**
     * @return the _dateTimes
     */
    public List<DateFormat> getDateTimes() {
        return _dateTimes;
    }

    /**
     * @return the _numbers
     */
    public List<NumberFormat> getNumbers() {
        return _numbers;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Class: " + getClass().getName());
        sb.append("\n  BIGDECIMAL_DIGITS_FROM = " + BIGDECIMAL_DIGITS_FROM);
        sb.append("\n  Cached BigDecimal constructor = " + (_bigDecimalConstructor != null));
        sb.append("\n  Locale = " + _locale);
        sb.append("\n  TimeZone = " + _timeZone);
        sb.append("\n  Decimal symbols:");
        sb.append("\n    getCurrencySymbol = " + decimalSymbols.getCurrencySymbol());
        sb.append("\n    getExponentSeparator = " + decimalSymbols.getExponentSeparator());
        sb.append("\n    getInfinity = " + decimalSymbols.getInfinity());
        sb.append("\n    getInternationalCurrencySymbol = " + decimalSymbols.getInternationalCurrencySymbol());
        sb.append("\n    getNaN = " + decimalSymbols.getNaN());
        sb.append("\n    getCurrency = " + decimalSymbols.getCurrency());
        sb.append("\n    getDecimalSeparator = " + decimalSymbols.getDecimalSeparator());
        sb.append("\n    getDigit = " + decimalSymbols.getDigit());
        sb.append("\n    getGroupingSeparator = " + decimalSymbols.getGroupingSeparator());
        sb.append("\n    getMinusSign = " + decimalSymbols.getMinusSign());
        sb.append("\n    getMonetaryDecimalSeparator = " + decimalSymbols.getMonetaryDecimalSeparator());
        sb.append("\n    getPatternSeparator = " + decimalSymbols.getPatternSeparator());
        sb.append("\n    getPerMill = " + decimalSymbols.getPerMill());
        sb.append("\n    getPercent = " + decimalSymbols.getPercent());
        sb.append("\n    getZeroDigit = " + decimalSymbols.getZeroDigit());
//    private DecimalFormatSymbols decimalSymbols;
        Date sample = new Date();
        sb.append("\n  Date formats:");
        for (int i = 0; i < _dates.size(); i++) {
            sb.append("\n    ");
            sb.append(_dates.get(i).format(sample));
        }
        sb.append("\n  Time formats:");
        for (int i = 0; i < _times.size(); i++) {
            sb.append("\n    ");
            sb.append(_times.get(i).format(sample));
        }
        sb.append("\n  DateTime formats:");
        for (int i = 0; i < _dateTimes.size(); i++) {
            sb.append("\n    ");
            sb.append(_dateTimes.get(i).format(sample));
        }
        Double number = 123456.789;
        sb.append("\n  Number formats:");
        for (int i = 0; i < _numbers.size(); i++) {
            NumberFormat nf = _numbers.get(i);
            sb.append("\n    ");
            sb.append(nf.format(number));
            sb.append("\n    isGroupingUsed  = ");
            sb.append(nf.isGroupingUsed());
            sb.append("\n    isParseIntegerOnly  = ");
            sb.append(nf.isParseIntegerOnly());
            sb.append("\n    getCurrency  = ");
            sb.append(nf.getCurrency());
            sb.append("\n    getMinimumIntegerDigits  = ");
            sb.append(nf.getMinimumIntegerDigits());
            sb.append("\n    getMaximumIntegerDigits  = ");
            sb.append(nf.getMaximumIntegerDigits());
            sb.append("\n    getMinimumFractionDigits  = ");
            sb.append(nf.getMinimumFractionDigits());
            sb.append("\n    getMaximumFractionDigits  = ");
            sb.append(nf.getMaximumFractionDigits());
            sb.append("\n    getRoundingMode  = " + nf.getRoundingMode());
        }

        return sb.toString();
    }
}
