/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.utils;

import java.util.HashMap;
import java.util.Map;

/**
 * Used to keep track of process (parsing or formatting) to avoid duplicates and
 * infinite loops.
 * 
 * @author ssg
 */
public class ObjectsRegistry {

    public static final String REFERENCE_DELIMITER = "#";
    public static final String REFERENCE_PREFIX = REFERENCE_DELIMITER + "REF" + REFERENCE_DELIMITER;
    private Map<Object, String> _processedObjects = new HashMap<Object, String>();
    private Map<String, Object> _processedKeys = new HashMap<String, Object>();
    private Map<String, Object> _referencedObjects = new HashMap<String, Object>();
    private Map<Object, String> _referenceObjects = new HashMap<Object, String>();
    private Map<Object, Object> _iid2Object = new HashMap<Object, Object>();

    /**
     * @return the _processedObjects
     */
    public Map<Object, String> getProcessedObjects() {
        return _processedObjects;
    }

    /**
     * @return the _referencedObjects
     */
    public Map<String, Object> getReferencedObjects() {
        return _referencedObjects;
    }

    /**
     * @return the _referenceObjects
     */
    public Map<Object, String> getReferenceObjects() {
        return _referenceObjects;
    }

    public void registerProcessedObject(String objectPath, Object object) {
        //System.out.println("registerProcessedObject: " + objectPath + "; " + ((object!=null) ? object.getClass().getName() : "null"));
        if (objectPath != null) {
            // TODO: check if path has registered object and remove ald registrations replacing with new identityHashCode for new object...

            if (getProcessedKeys().containsKey(objectPath)) {
                Object uoid = getProcessedKeys().get(objectPath);
                if (uoid != null && uoid.equals(System.identityHashCode(object))) {
                } else {
                    if (_processedObjects.containsKey(System.identityHashCode(object))) {
                        _processedObjects.remove(System.identityHashCode(object));
                    }
                }
            }
            _processedObjects.put(System.identityHashCode(object), objectPath);
            getProcessedKeys().put(objectPath, object);
            _iid2Object.put(System.identityHashCode(object), object);
        }
    }

    public String getProcessedObjectPath(Object object) {
        if (isProcessedObject(object)) {
            return getProcessedObjects().get(System.identityHashCode(object)).toString();
        } else {
            return null;
        }
    }

    public boolean isProcessedObject(Object object) {
        return object != null && _processedObjects.containsKey(System.identityHashCode(object));
    }

    public boolean isReferencable(Object object) {
        return !Utilities.isScalarOrString(object);
    }

    public void registerReferencedObject(String objectPath, Object object) {
        // TODO: update all occurences of path
        if (objectPath != null) {
            getReferencedObjects().put(objectPath, object);
        }
    }

    public void registerReferenceObject(Object object, String referencePath) {
        // TODO: update all occurences of path
        if (referencePath != null) {
            getReferenceObjects().put(object, referencePath);
        }
    }

    public boolean isReference(Object object) {
        return object instanceof String && ((String) object).startsWith(REFERENCE_PREFIX);
    }

    public Object resolveReference(Object reference) {
        //System.out.println("resolveReference: [" + (reference instanceof String && ((String) reference).startsWith(REFERENCE_PREFIX)) + ", " + isProcessedObject(reference) + "] for " + reference);
        if (isReference(reference)) {// && isProcessedObject(reference)) {
            // 0-""#1-"REF"#2-"className"#3-"path"
            String[] values = ((String) reference).split(REFERENCE_DELIMITER);
            Object obj = getReferencedObjects().get(values[3]);
            if (obj == null) {
                obj = getProcessedKeys().get(values[3]);
            }
            return obj;
        } else {
            return reference;
        }
    }

    /**
     * @return the _processedKeys
     */
    public Map<String, Object> getProcessedKeys() {
        return _processedKeys;
    }

    public static String buildReferenceString(Class clazz, String path) {
        return REFERENCE_PREFIX + ((clazz != null) ? clazz.getCanonicalName() : "null") + REFERENCE_DELIMITER + path;
    }
}
