/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.utils;

import java.util.ArrayList;
import java.util.List;

/**
 * Abstracts path to simplify navigation and analysis usually when parsing.
 * Provides a number of utility methods to test and build paths.
 *
 * @author ssg
 */
public class Path {

    public static final String SEPARATOR = "/";
    public static final String ARRAY_OPEN = "[";
    public static final String ARRAY_CLOSE = "]";
    public static final int NOT_INDEXED = -1;
    public static final int NON_NUMERIC_INDEX = -2;
    List<String> _path = new ArrayList<String>();
    int _depth;

    private Path() {
    }

    public Path(String path) {
        add(path);
    }

    public Path(String path, int index) {
        add(path, index);
    }

    public Path(Path parent, String path, int index) {
        copyPath(parent);
        add(path, index);
    }

    public void add(String path, int index) {
        add(path, (index > NOT_INDEXED) ? "" + index : null);
    }

    public void add(String path, String index) {
        if (SEPARATOR.equals(path) || path == null || path.equals("")) {
            return;
        }
        _depth++;
        _path.add(SEPARATOR);
        _path.add(path);
        if (index != null) {
            _path.add(ARRAY_OPEN);
            _path.add(index);
            _path.add(ARRAY_CLOSE);
        }
        //System.out.println("PATH: " + toString());
        //System.out.println("... added      : " + toString()+". TAG: "+path);
    }

    public void add(String path) {
        add(path, NOT_INDEXED);
    }

    public String remove() {
        //System.out.println("... remove from: " + toString());
        if (_path.size() > 0) {
            String last = null;
            if (isArray()) {
                _path.remove(_path.size() - 1);
                _path.remove(_path.size() - 1);
                _path.remove(_path.size() - 1);
            }
            if (_path.size() > 0) {
                last = _path.remove(_path.size() - 1);
                if (_path.size() > 0 && SEPARATOR.equals(_path.get(_path.size() - 1))) {
                    _path.remove(_path.size() - 1);
                }
            }
            _depth--;
            //System.out.println("... removed    : " + toString() + ". REMOVED: " + last);
            return last;
        }
        return null;
    }

    public void clear() {
        _path.clear();
        _depth = 0;
    }

    @Override
    public String toString() {
        return toString(false);
    }

    public String toString(boolean noArrays) {
        StringBuilder sb = new StringBuilder();
        int skip = 0;
        for (String s : _path) {
            if (noArrays && ARRAY_OPEN.equals(s)) {
                skip = 3;
            }
            if (skip == 0) {
                sb.append(s);
            }
            if (skip > 0) {
                skip--;
            }
        }
        return sb.toString();
    }

    /**
     * Returns path object for specified level.
     * If level is negative or exceeds path depth then null is returned.
     */
    public Path getPathForLevel(int level) {
        if (level < 0 || level > this.getDepth()) {
            return null;
        }
        Path p = new Path(null);
        p.copyPath(this);
        while (p.getDepth() > level) {
            p.remove();
        }
        return p;
    }

    public Path getParent() {
        if (getDepth() == 0) {
            return null;
        }
        Path result = new Path();
        result.copyPath(this);
        result.remove();
        return result;
    }

    public void copyPath(Path from) {
        for (int i = 0; i < from._path.size(); i++) {
            _path.add(from._path.get(i));
        }
        _depth += from.getDepth();
    }

    public boolean isArray() {
        if (_path.isEmpty() || _path.size() < 3) {
            return false;
        }
        if (ARRAY_CLOSE.equals(_path.get(_path.size() - 1)) && ARRAY_OPEN.equals(_path.get(_path.size() - 3))) {
            return true;
        } else {
            return false;
        }
    }

    public int getIndexInt() {
        if (isArray()) {
            try {
                return Integer.parseInt(getIndexString());
            } catch (NumberFormatException nfex) {
                return NON_NUMERIC_INDEX;
            }
        } else {
            return NOT_INDEXED;
        }
    }

    public String getIndexString() {
        if (isArray()) {
            return _path.get(_path.size() - 2);
        } else {
            return null;
        }
    }

    public void setIndex(int index, boolean enforce) {
        if (index == NOT_INDEXED) {
            setIndex(null, enforce);
        } else {
            setIndex("" + index, enforce);
        }
    }

    public void setIndex(String index, boolean enforce) {
        if (isArray() || enforce) {
            if (isArray()) {
                if (index == null) {
                    _path.remove(_path.size() - 1);
                    _path.remove(_path.size() - 1);
                    _path.remove(_path.size() - 1);
                } else {
                    _path.set(_path.size() - 2, index);
                }
            } else if (index != null) {
                _path.add(ARRAY_OPEN);
                _path.add(index);
                _path.add(ARRAY_CLOSE);
            }
        }
    }

    public String getLeafName() {
        if (_path.isEmpty()) {
            return "";
        }
        if (isArray()) {
            return _path.get(_path.size() - 4);
        } else {
            return _path.get(_path.size() - 1);
        }
    }

    public int getDepth() {
        return _depth;
    }

    public boolean isRoot() {
        return _path.isEmpty();
    }

    public static Path parse(String path) {
        Path p = new Path();
        if (path != null) {
            String[] items = path.split(SEPARATOR);
            for (String item : items) {
                if (item.endsWith(ARRAY_CLOSE) && item.indexOf(ARRAY_OPEN) > 0) {
                    int idx = item.indexOf(ARRAY_OPEN);
                    p.add(item.substring(0, idx), item.substring(idx + 1, item.length() - 1));
                } else {
                    if (item.length() > 0) {
                        p.add(item);
                    }
                }
            }
        }
        return p;
    }

    public static Path createFor(Path parent, String subpath, String index) {
        Path path = new Path();
        if (parent != null) {
            path.copyPath(parent);
        }
        path.add(subpath, index);
        return path;
    }
}
