/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.utils;

import com.ssg.tools.jsonxml.utils.Utilities.Pair;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * Reflective builder converts List/Map hierarchical structures into objects
 * using list of registered class mappings and reflection (getters/setters only).
 * 
 * @author ssg
 */
public class ReflectiveBuilder {

    public static String REFLECTION_COMPLEX_MAP_KEY = "RCMK_84c17cc5_64ed_4645_bd96_907ab824c653";

    /**
     * This class is used to wrap complex maps' entries.
     * @param <A>
     * @param <B> 
     */
    public static class PairForReflection<A, B> extends Pair<A, B> {

        public PairForReflection(A a, B b) {
            super(a, b);
        }

        public String getAType() {
            return (getA() != null) ? getA().getClass().getName() : null;
        }

        public String getBType() {
            return (getB() != null) ? getB().getClass().getName() : null;
        }
    }
    List<Class> _classes = new ArrayList<Class>();
    Map<String, Class> _mappedClasses = new LinkedHashMap<String, Class>();

    public ReflectiveBuilder() {
    }

    public boolean registerClass(Class clazz) {
        if (clazz != null) {
            if (_classes.contains(clazz)) {
                return true;
            }
            try {
                Constructor cs = clazz.getConstructor(null);
                _classes.add(clazz);
                return true;
            } catch (Throwable th) {
                return false;
            }
        } else {
            return false;
        }
    }

    public boolean mapClass(String name, Class clazz) {
        if (registerClass(clazz)) {
            _mappedClasses.put(name, clazz);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Override to add/remove non-reflectable classes.
     * Return false if class should not be processed reflectively.
     * Currently excludes from reflective handling all java.* and sun.* classes.
     *
     * @param object
     * @return
     */
    boolean testIsReflectable(Object object) {
        if (object == null) {
            return false;
        }
        String className = object.getClass().getName();
        return !(className.startsWith("java.") || className.startsWith("sun."));
    }

    /**
     * Builds map of name=value pairs for object using reflection.
     * Only getters (get* without parameters) are used to do it.
     *
     * @param object
     * @return
     */
    public Map<String, Object> getReflectable(Object object) {
        Map obj2map = new LinkedHashMap();
        if (!testIsReflectable(object)) {
            // ??? we do not reflect java standard classes... -> as strings!
        } else {
            // use java and groovy reflection for arbitrary types...
            Method[] ms = object.getClass().getMethods();
            for (Method m : ms) {
                if (!m.getName().equals("getClass") && m.getName().startsWith("get") && m.getParameterTypes().length == 0 && m.getReturnType() != void.class) {
                    try {
                        Object v = m.invoke(object);
                        obj2map.put(m.getName().substring(3), (v != null) ? v : null);
                    } catch (Throwable th) {
                    }
                }
            }
        }
        return obj2map;
    }

    /**
     * Rebuild object reflectively and replace in parent.
     * 
     * @param parent
     * @param object
     * @param path
     * @param formats
     * @throws StructureProcessingException 
     */
    public void reflectiveBuild(Object parent, Object object, Path path, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        String name = path.getLeafName();
        Object object2 = reflectiveBuild(object, path, formats, objectsRegistry);
        //System.out.println("reflectiveBuild at "+path+": "+object+" -> "+object2);
        if (parent != null && object2 != null && !object2.equals(object)) {
            if (parent instanceof List) {
                List list = (List) parent;
                list.set(list.indexOf(object), object2);
            } else if (parent instanceof Map) {
                Map map = (Map) parent;
                map.put(name, object2);
            } else {
                String mn = "set" + name;
                Method[] ms = parent.getClass().getMethods();
                for (Method m : ms) {
                    if (m.getName().equals(mn) && m.getParameterTypes().length == 1 && m.getParameterTypes()[0].isAssignableFrom(object2.getClass())) {
                        try {
                            m.invoke(parent, object2);
                            break;
                        } catch (Throwable th) {
                            throw new StructureProcessingException("Failed to set " + name + " for " + parent + " with " + object2, th);
                        }
                    }
                } // END: iterate methods
            }
        }
    }

    public Object reflectiveBuild(Object object, Path basePath, Formats formats, ObjectsRegistry objectsRegistry) {
        if (object == null || basePath == null) {
            return object;
        }
        Path path = new Path(null);
        path.copyPath(basePath);
        String name = path.getLeafName();
        if (_mappedClasses.containsKey(name)) {
            Object object2 = null;
            try {
                object2 = reflectiveBuild(object, _mappedClasses.get(name), path, formats, objectsRegistry);
            } catch (Throwable th) {
                th.printStackTrace(System.out);
            }
            return (object2 != null) ? object2 : object;
        } else {
            return object;
        }
    }

    /**
     * Recursively builds object for specified class from currently available context objects.
     *
     * @param context
     * @param level
     * @param name
     * @param object
     * @param clazz
     * @return
     * @throws StructureProcessingException
     */
    Object reflectiveBuild(Object object, Class clazz, Path path, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        //System.out.println("reflectiveBuild: " + path + ": " + clazz + ": " + object);
        if (clazz == null || object == null || !(object instanceof Map)) {
            if (Utilities.isNumericClass(clazz)) {
                Object result = formats.asNumber(object, clazz);
                if (result != null) {
                    return result;
                }
            }
            return object;
        }
        Map values = (Map) object;
        Object result = null;
        // Instantiate object. If failure then find 1st compatible
        // class in registered classes list and try it.
        try {
            result = clazz.newInstance();
            //System.out.println("  instantiated " + clazz.getName());
        } catch (Throwable th) {
            if (clazz.equals(List.class)) {
                clazz = ArrayList.class;
            } else if (clazz.equals(Map.class)) {
                clazz = LinkedHashMap.class;
            } else {
                for (Class cl : _classes) {
                    if (clazz.isAssignableFrom(cl)) {
                        clazz = cl;
                    }
                }
            }
            if (clazz != null) {
                try {
                    result = clazz.newInstance();
                    //System.out.println("  instantiated " + clazz.getName());
                } catch (Throwable th1) {
                    throw new StructureProcessingException("Failed to create instance of object " + clazz + ".", th1);
                }
            }
        }

        // test for COMPLEX MAP...
        if (isWrappedMap(values)) {
            return unwrapMap(values, path, formats, objectsRegistry);
        }

        // we got here because of normal reflection
        for (Method m : clazz.getMethods()) {
            if (m.getName().startsWith("set") && m.getParameterTypes().length == 1) {
                String key = m.getName().substring(3);
                try {
                    path.add(key);
                    Class paramType = m.getParameterTypes()[0];
                    Object value = values.get(key);

                    if (value == null) {
                        value = values.get(key.toLowerCase());
                        if (value == null) {
                            value = values.get(key.toUpperCase());
                        }
                    }
                    // pre-process "null" string
                    if (value != null && value.equals("null")) {
                        // && !paramType.equals(String.class)) {
                        value = null;
                    }
                    if (paramType.isEnum() && value instanceof String) {
                        value = PRIMITIVE_TYPE.valueOf(paramType, (String) value);
                    } else if (value != null && !paramType.isAssignableFrom(value.getClass())) {
                        value = reflectiveBuild(value, paramType, path, formats, objectsRegistry);
                        //System.out.println("  converted " + value + " to " + paramType.getName());
                    } else if (value instanceof Map && ((Map) value).size() == 1 && ((Map) value).containsKey(REFLECTION_COMPLEX_MAP_KEY)) {
                        // complex map restore...
                        value = reflectiveBuild(value, paramType, path, formats, objectsRegistry);
                    }
                    try {
                        m.invoke(result, value);
                    } catch (Throwable th) {
                        throw new StructureProcessingException("Failed to invoke " + m.getName() + " for " + object + " with " + value, th);
                    }
                } finally {
                    path.remove();
                }
            }
        }

        objectsRegistry.registerProcessedObject(path.toString(), result);
        return result;
    }

    /**
     * Detects if map is wrapped...
     * 
     * @param values
     * @return 
     */
    public static boolean isWrappedMap(Map values) {
        return values != null && values.containsKey(REFLECTION_COMPLEX_MAP_KEY);
    }

    /**
     * Wraps map if needed to allow keys as objects (not just strings)
     * 
     * @param map
     * @return 
     */
    public static Map wrapMap(Map map) {
        Collection keys = map.keySet();
        boolean stringKeysOnly = true;
        for (Object key : keys) {
            if (!(key instanceof String)) {
                stringKeysOnly = false;
                break;
            }
        }

        if (!stringKeysOnly) {
            // proceed with EntrySet: create replacement map and update keys!!! really bad hack...
            Map newMap = new LinkedHashMap();
            List entriesList = new ArrayList();
            for (Object obj : map.entrySet()) {
                entriesList.add(new PairForReflection(((Entry) obj).getKey(), ((Entry) obj).getValue()));
            }
            newMap.put(REFLECTION_COMPLEX_MAP_KEY, entriesList);
            map = newMap;
        }

        return map;
    }

    /**
     * Performs opposite operation to wrapMap
     * 
     * @param values
     * @param path
     * @param formats
     * @return 
     */
    public Map unwrapMap(Map values, Path path, Formats formats, ObjectsRegistry objectsRegistry) {
        // test for COMPLEX MAP...
        //if (values != null && values.size() == 1 && values.containsKey(Utilities.REFLECTION_COMPLEX_MAP_KEY)) {
        if (isWrappedMap(values)) {
            List pairs = null;
            if (true) {
                Object obj = values.get(REFLECTION_COMPLEX_MAP_KEY);
                if (obj instanceof List) {
                    pairs = (List) obj;
                } else if (obj instanceof Map && ((Map) obj).size() == 1) {
                    pairs = new ArrayList();
                    pairs.add(((Map) obj).values().iterator().next());
                } else {
                    // we failed to restore complex map.... :(
                    return values;
                }
            }

            Map result = new LinkedHashMap();
            ClassLoader classLoader = Thread.currentThread().getContextClassLoader();

            // iterate pairs and try to build nested named A and B objects
            Path p = path.createFor(path, REFLECTION_COMPLEX_MAP_KEY, null);
            int idx = 0;
            try {
                for (Object obj : pairs) {
                    p.setIndex("" + idx, true);
                    Map map = (Map) obj;
                    // key type
                    Class classA = (objectsRegistry.isReference(map.get("AType"))) ? (Class) objectsRegistry.resolveReference(map.get("AType")) : classLoader.loadClass((String) map.get("AType"));
                    objectsRegistry.registerProcessedObject(p.createFor(p, "AType", null).toString(), classA);

                    // value type
                    //Class classB = classLoader.loadClass((String) objectsRegistry.resolveReference(map.get("BType")));
                    Class classB = (objectsRegistry.isReference(map.get("BType"))) ? (Class) objectsRegistry.resolveReference(map.get("BType")) : classLoader.loadClass((String) map.get("BType"));
                    objectsRegistry.registerProcessedObject(p.createFor(p, "BType", null).toString(), classB);

                    // key object
                    Object key = (objectsRegistry.isReference(map.get("A")))
                            ? objectsRegistry.resolveReference(map.get("A"))
                            : reflectiveBuild(map.get("A"), classA, path, formats, objectsRegistry);
                    objectsRegistry.registerProcessedObject(p.createFor(p, "A", null).toString(), key);

                    // value object
                    Object val = (objectsRegistry.isReference(map.get("B")))
                            ? objectsRegistry.resolveReference(map.get("B"))
                            : reflectiveBuild(map.get("B"), classA, path, formats, objectsRegistry);
                    ((Map) result).put(key, val);
                    objectsRegistry.registerProcessedObject(p.createFor(p, "B", null).toString(), val);

                    idx++;
                }
            } catch (Throwable th) {
                // for any error return original object
                th.printStackTrace(System.out);
                return values;
            }
            return result;
        }
        return values;
    }
}
