/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.utils;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.math.BigDecimal;
import java.util.Date;

/**
 * Shared methods
 * 
 * @author ssg
 */
public class Utilities {

    public static class Pair<A, B> {

        private A a;
        private B b;

        public Pair(A a, B b) {
            setA(a);
            setB(b);
        }

        /**
         * @return the a
         */
        public A getA() {
            return a;
        }

        /**
         * @param a the a to set
         */
        public void setA(A a) {
            this.a = a;
        }

        /**
         * @return the b
         */
        public B getB() {
            return b;
        }

        /**
         * @param b the b to set
         */
        public void setB(B b) {
            this.b = b;
        }
    }

    /**
     * converts noun from single form into multiple as follows (samples):
     *   offer 2 offers
     *   business 2 businesses
     *   binary 2 binaries
     *
     * @param key
     * @return
     */
    public static String single2multiple(String key) {
        if (key.endsWith("s") && !key.endsWith("ss")) {
            return key;
        } else if (key.endsWith("ss")) {
            return key + "es";
        } else if (key.endsWith("y")) {
            return key.substring(0, key.length() - 1) + "ies";
        } else {
            return key + "s";
        }
    }

    /**
     * restores noun from multiple into single form reversing
     * actions applied in single2multiple as follows (samples):
     *   offers 2 offer
     *   businesses 2 business
     *   binaries 2 binary
     *
     * @param key
     * @return
     */
    public static String multiple2single(String key) {
        if (key.endsWith("ies")) {
            return key.substring(0, key.length() - 3) + 'y';
        } else if (key.endsWith("ses")) {
            return key.substring(0, key.length() - 2);
        } else if (key.endsWith("s")) {
            return key.substring(0, key.length() - 1);
        } else {
            return key;
        }
    }

    public static boolean isNumericClass(Class clazz) {
        return clazz.equals(Byte.class)
                || clazz.equals(byte.class)
                || clazz.equals(Short.class)
                || clazz.equals(short.class)
                || clazz.equals(Integer.class)
                || clazz.equals(int.class)
                || clazz.equals(Long.class)
                || clazz.equals(long.class)
                || clazz.equals(Float.class)
                || clazz.equals(float.class)
                || clazz.equals(Double.class)
                || clazz.equals(double.class)
                || clazz.equals(BigDecimal.class);
    }

    public static boolean isScalarOrString(Object object) {
        if (object == null) {
            return true;
        }
        Class c = object.getClass();
        if (c.isPrimitive() || c.isEnum()) {
            return true;
        }
        if (object instanceof Boolean) {
            return true;
        }
        if (object instanceof Character) {
            return true;
        }
        if (object instanceof String) {
            return true;
        }
        if (object instanceof Date) {
            return true;
        }
        if (isNumericClass(object.getClass())) {
            return true;
        }
        return false;
    }

    public static Reader getReaderForResource(String name) throws IOException {
        try {
            // works in unit tests mode
            return new InputStreamReader(Thread.currentThread().getContextClassLoader().getResourceAsStream(name));
        } catch (NullPointerException npex) {
            // works in application test mode
            return new FileReader(new File(new File("./src/test/java"), name));
        }
    }
}
